@@ -5,6 +5,7 @@ import os
 
 import requests
 import pandas as pd
+import numpy as np
 from io import BytesIO, StringIO
 
 class Client(object):
@@ -62,7 +63,7 @@ class Client(object):
 
     # DESCRIPTORS
 
-    def get_descriptors(self,substance_names,descriptor_set='all'):
+    def get_descriptors(self,substance_names,descriptor_set='all',include_properties=False):
         """Get substance descriptors by providing substance names.
 
         Descriptors are looked up and returned for all substance names
@@ -87,7 +88,8 @@ class Client(object):
             that are real-valued across most available substances.
             To see all available descriptor sets, log in by web browser
             and examine the featurization step of a modeling pipeline.
-            
+        include_properties : bool
+            Flag for including substance properties in the featurization.    
 
         Returns
         -------
@@ -106,7 +108,8 @@ class Client(object):
             json={
                 'substance_names':substance_names,
                 'descriptor_set':descriptor_set,
-                'process_id':procid
+                'process_id':procid,
+                'include_properties':include_properties
             }
         )
         if resp.status_code == 200:
@@ -132,8 +135,8 @@ class Client(object):
         ----------
         featurization : str
             Specifies which featurization to apply. 
-            This should be 'lithium_ionic', 'transition_metal_redox', or 'molecular'.
-            The 'molecular' and 'transition_metal_redox' options are disabled by default 
+            This should be 'lithium_ionic', 'anion_redox', or 'molecular'.
+            The 'molecular' and 'anion_redox' options are disabled by default 
             (only specific instances of the application are allowed to use them).
         file_paths : dict 
             Dict of file paths containing substance data.
@@ -216,10 +219,15 @@ class Client(object):
         """
         if wait_interval < 1: wait_interval = 1.
         # check the process
-        procstat = self.session.get(
+        resp = self.session.get(
             self.service_url+'process_status',
             params={'process_id':process_id}
-        ).json()
+        )
+        if resp.status_code == 200: 
+            procstat = resp.json()
+        else:
+            procstat = {'status':'ERROR','status_code':resp.status_code} 
+        n_waits = 0
         if wait:
             while not procstat['status'] in ['FINISHED','ERROR']: 
                 time.sleep(wait_interval)
@@ -227,6 +235,10 @@ class Client(object):
                     self.service_url+'process_status',
                     params={'process_id':process_id}
                 ).json()
+                n_waits += 1
+                # gradually slow down status checks
+                if np.mod(n_waits,100) == 0 and wait_interval < 30:
+                    wait_interval += 1
         if procstat['status'] == 'FINISHED':
             # NOTE: if it was a fractionally occupied crystal,
             # and the job was submitted with return_files==True,
@@ -1107,6 +1119,8 @@ class Client(object):
 
         Parameters
         ----------
+        model_id : int
+            Integer id of the model being applied.
         process_id : int
             Integer id, as assigned to the process when it was started
         wait : bool
@@ -1189,7 +1203,7 @@ class Client(object):
         else:
             return {'success':False,'status_code':resp.status_code}
 
-    def configure_candidates(self,design_id,substance_settings={},scalar_settings={},model_settings={}):
+    def configure_candidates(self,design_id,settings={},substance_settings={},scalar_settings={},model_settings={}):
         """Perform candidate configuration: step 2 of the design pipeline.
 
         The goal of this step is to specify 
@@ -1199,6 +1213,12 @@ class Client(object):
         ----------
         design_id : int
             Integer id of the design being worked on.
+        settings : dict
+            Dict of general candidate configurations:
+
+            - 'candidate_source' (str): Specifies source of candidates. Either 'candidate_generator' (default) or 'dataset'.
+            - 'candidate_dataset_id' (int): If 'candidate_source'=='dataset', this is the integer id of the dataset that defines the candidates.
+
         substance_settings : dict
             Settings for generating candidate substances.
             The keys of this dict should be the names of the substance columns.
@@ -1230,6 +1250,7 @@ class Client(object):
             Each value must be the name of the model to use to estimate values for the column.
         """
         data = {'design_id':design_id}
+        data.update(settings)
         for subsnm,stgs in substance_settings.items():
             for stgnm,stgval in stgs.items():
                 data[subsnm+'_'+stgnm] = stgval
@@ -1317,6 +1338,33 @@ class Client(object):
 
     # SUBSTANCES
 
+    def check_substance_names(self,names,location='shared'):
+        """Check for existence of substances by name.
+
+        Parameters
+        ----------
+        names : list
+            List of names (strings) to be checked.
+        location : str
+            Location of substances ('private' or 'shared').
+
+        Returns
+        -------
+        response : dict
+            Dict containing status report and results.
+        """
+        resp = self.session.get(
+            self.service_url+'api/check_substance_names', 
+            headers={'x-api-key':self.api_key},
+            json={'names':names,'location':location}
+        )
+        if resp.status_code == 200:
+            respdata = resp.json()
+            return respdata
+        else:
+            return {'success':False,'status_code':resp.status_code}
+
+
     def pull_shared_substance(self,name):
         """Copy a substance from the shared to the private substance tables.
 
@@ -1391,18 +1439,263 @@ class Client(object):
         else:
             return {'success':False,'status_code':resp.status_code}
 
-    def get_substance_data(self,name):
+    def get_substance_data(self,name,location='shared'):
         """Get all data for a substance by providing a substance name.
 
-        See the documentation for
-        apyonics.substances_client.SubstancesClient.get_substance_data().
+        Parameters
+        ----------
+        name : str 
+            Name of the substance.
+        location : str
+            Location of the substance ('private' or 'shared'). 
+
+        Returns
+        -------
+        resp : dict
+            Contains status report and, if successful, substance data. 
         """
         resp = self.session.get(
             self.service_url+'api/download_substance', 
-            json={'name':name}
+            json={'name':name,'location':location}
+        )
+        if resp.status_code == 200:
+            respdata = {'success':True,'data':resp.json()}
+            return resp.json()
+        else:
+            return {'success':False,'status_code':resp.status_code}
+
+
+    # MEASUREMENTS
+
+    def get_measurement_index(self):
+        """Download the index of available measurements.
+
+        Returns
+        -------
+        response : dict
+            Dict containing information about measurements available to the client.
+        """
+        resp = self.session.get(self.service_url+'api/measurement_index')
+        if resp.status_code == 200:
+            return resp.json()
+        else:
+            return {'success':False,'status_code':resp.status_code}
+
+    def new_measurement(self,filepath,type,name,description=''):
+        """Create a new measurement on Aionics.
+
+        The data for the measurement may be in any format,
+        provided there is a processing workflow available on platform for it.
+
+        Parameters
+        ----------
+        filepath : str
+            Path to measurement source file on the local filesystem.
+        type : str
+            String describing the measurement type. 
+            Valid types:
+
+            - 'cycler_nda'
+
+        name : str
+            Name for the new measurement- must be unique.
+        description : str
+            Description of the new measurement (optional).
+
+        Returns
+        -------
+        response : dict
+            Dict containing the new measurement id or an error report.
+        """
+        resp = self.session.post(
+            self.service_url+'api/new_measurement',
+            files={
+                'source_file':open(filepath,'rb'),
+                'metadata':json.dumps({
+                    'type':type,
+                    'name':name,
+                    'description':description
+                }).encode()
+            }
+        )
+        if resp.status_code == 200:
+            return resp.json()
+        else:
+            return {'success':False,'status_code':resp.status_code}
+
+    def process_measurement(self,measurement_id,settings):
+        """Start processing a measurement.
+
+        Parameters
+        ----------
+        measurement_id : int
+            Integer id of the measurement to be edited.
+        settings : dict
+            Dict of measurement processing configurations.
+            Configurations include:
+
+            - 'processing': Selects a processing workflow. Options are: 
+
+                - 'cycle_summary_v1': For 'cycler_nda' type measurements, process cycle data.
+                - 'cycle_life_rpt_hppc_v1': For 'cycler_nda' type measurements, process cycle data, determine cycle life, and interpret the RPT and HPPC metrics of selected cycles.
+
+            - 'featurization': Selects a processing workflow. Options are:
+
+                - 'discharge_capacity': For 'cycler_nda' type measurements, apply the featurization of Severson and Attia, et al, for predicting cycle life from early cycle data.
+
+            - For 'cycler_nda' type measurements to be processed by 'cycle_summary_v1' or 'cycle_life_rpt_hppc_v1', the following settings are required:
+
+                - 'active_material_mass': The mass of the active material, in grams, for determining specific capacity.
+                - 'cathode_area': Real-valued area of a single layer of cathode material, for determining ASI.
+                - 'layer_count': Integer number of cathode layers, for determining ASI.
+                - 'ignore_cycles': List of cycle id's that should be ignored for featurization, cycle summary, and determination of cycle life.
+
+            - For 'cycler_nda' type measurements to be processed by 'cycle_life_rpt_hppc_v1', the following settings are required:
+
+                - 'cycle_life_threshold': Percentage of initial cycle life at which the battery should be considered dead.
+                - 'rpt_cycles': List of cycle id's that should be processed as RPT tests.
+                - 'hppc_cycles': List of cycle id's that should be processed as HPPC tests.
+
+            - For 'cycler_nda' type measurements to be featurized by 'discharge_capacity', the following settings are required:
+
+                - 'qofv_index_1': Cycle id for the first Q(V) profile 
+                - 'qofv_index_2': Cycle id for the second Q(V) profile 
+                - 'capacity_index_1': Cycle id for the first Q(N) point for capacity trajectory fitting 
+                - 'capacity_index_2': Cycle id for the intermediate Q(N) point for capacity trajectory fitting 
+                - 'capacity_index_3': Cycle id for the final Q(N) point for capacity trajectory fitting 
+
+        Returns
+        -------
+        resp : dict
+            Contains status report and, if successful,
+            a process id for monitoring the process.
+        """
+        pid_resp = self.session.get(
+            self.service_url+'measurement_process_id', 
+            json={'measurement_id':measurement_id}
+        )
+        if not pid_resp.status_code == 200:
+            return {'success':False,'status_code':pid_resp.status_code}
+        procid = pid_resp.json()['process_id']
+        data = settings.copy()
+        data.update({
+            'measurement_id':measurement_id,
+            'process_id':procid
+        })
+        resp = self.session.post(
+            self.service_url+'process_measurement', 
+            json=data
+        )
+        if resp.status_code == 200:
+            return resp.json()
+        else:
+            return {'success':False,'status_code':resp.status_code}
+
+    def get_measurement_data(self,measurement_id):
+        """Get measurement settings and results.
+
+        Parameters
+        ----------
+        measurement_id : int
+            Integer id of the measurement to be queried.
+
+        Returns
+        -------
+        resp : dict
+            Contains status report and, if successful, 
+            the settings and results of the measurement.
+        """
+        settings_resp = self.session.get(
+            self.service_url+'measurement_settings', 
+            params={'measurement_id':measurement_id}
+        )
+        if settings_resp.status_code == 200:
+            settings = settings_resp.json()['data']
+        else:
+            return {'success':False,'status_code':resp.status_code}
+        results_resp = self.session.get(
+            self.service_url+'measurement_results', 
+            params={'measurement_id':measurement_id}
+        )
+        if results_resp.status_code == 200:
+            results = results_resp.json()['data']
+        else:
+            return {'success':False,'status_code':resp.status_code}
+        return {'success':True,'settings':settings,'results':results}
+
+    def download_measurement_file(self,measurement_id,file_tag,output_path):
+        """Download a measurement file by providing the file tag and measurement id.
+
+        Parameters
+        ----------
+        measurement_id : int
+            Integer index of the desired measurement.
+        file_tag : str
+            String indicating which file to download.
+            All measurements have a file with tag 'SOURCE_DATA'.
+            Depending on processing settings,
+            measurements of 'cycler_nda' type may also have:
+
+            - 'SOURCE_TABLE': .csv table unpacked from source binary
+            - 'STEP_TABLE': .csv table of key values from each cycler step
+            - 'CYCLE_TABLE': .csv table of key outcomes from each cycle
+            - 'RPT_TABLE': .csv table of RPT test results
+            - 'HPPC_TABLE': .csv table of HPPC test results
+
+        output_path : str
+            Local filesystem path where the file should be saved.
+
+        Returns
+        -------
+        response : dict 
+            Response data or error report. 
+        """
+        resp = self.session.get(
+            self.service_url+'measurement_file',
+            params={
+                'measurement_id':measurement_id,
+                'file_tag':file_tag
+            }
+        )
+        if resp.status_code == 200:
+            output_dir, fname = os.path.split(output_path)
+            if not os.path.exists(output_dir):
+                raise IOError('output directory {} does not exist'.format(output_dir))
+            open(output_path,'wb').write(resp._content)
+            return {'success':True}
+        else:
+            respdata = resp.json()
+            return {'success':False,'status_code':resp.status_code,'error':respdata['error']}
+
+    def edit_measurement_metadata(self,measurement_id,name='',description=''):
+        """Edit measurement metadata (name and description).
+
+        Parameters
+        ----------
+        measurement_id : int
+            Integer id of the measurement to be edited.
+        name : str
+            New name for the measurement being edited 
+            (must be unique among measurements).
+            If not provided, the name is not updated.
+        description : str
+            New description for the measurement being edited.
+            If not provided, the description is not updated.
+
+        Returns
+        -------
+        resp : dict
+            Contains status report.
+        """
+        resp = self.session.post(
+            self.service_url+'api/edit_measurement', 
+            json={
+                'measurement_id':measurement_id,
+                'name':name,
+                'description':description
+            }
         )
         if resp.status_code == 200:
-            respdata = {'success':True,'data':resp.json()['data']}
             return resp.json()
         else:
             return {'success':False,'status_code':resp.status_code}
