@@ -180,10 +180,10 @@ class SubstancesClient(object):
         response : dict
             Dict containing status report and results.
         """
-        resp = self.session.post(
+        resp = self.session.get(
             self.service_url+'check_substance_names', 
             headers={'x-api-key':self.api_key},
-            json={'substance_names':names}
+            json={'names':names}
         )
         if resp.status_code == 200:
             respdata = resp.json()
